/**
 * Copyright (c) 2010 jolira. All rights reserved. This program and the accompanying materials are made available under
 * the terms of the GNU Public License 2.0 which is available at http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 */

package com.google.code.joliratools.logdb.parser;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.StringTokenizer;

/**
 * A parser for processing delimited files.
 * 
 * @author jfk
 * @date Sep 13, 2010 11:09:15 PM
 * @since 1.0
 */
public class DelimitedParser implements Parser {
    private static int max(final int... vals) {
        int max = vals[0];

        for (int idx = 1; idx < vals.length; idx++) {
            if (vals[idx] > max) {
                max = vals[idx];
            }
        }

        return max;
    }

    private final String fieldDelimiters;
    private final int timeStampPos;
    private final String timeStampFormat;
    private final int maxPos;
    private final int classPos;
    private final int methodPos;
    private final int levelPos;

    /**
     * @param fieldDelimiters
     *            the delimiters to be used to break the records into fields
     * @param timeStampPos
     *            the position of the timestamp field
     * @param timeStampFormat
     *            the format of the timestamp field
     * @param classPos
     *            the position of the class name
     * @param methodPos
     *            the position of the method name
     * @param levelPos
     *            the position of the field containing the level
     */
    public DelimitedParser(final String fieldDelimiters, final int timeStampPos, final String timeStampFormat,
            final int classPos, final int methodPos, final int levelPos) {
        this.fieldDelimiters = fieldDelimiters;
        this.timeStampFormat = timeStampFormat;
        this.timeStampPos = timeStampPos;
        this.classPos = classPos;
        this.methodPos = methodPos;
        this.levelPos = levelPos;

        maxPos = max(timeStampPos, classPos, methodPos, levelPos);
    }

    private String getField(final String[] fields, final int pos) {
        if (pos < 0) {
            return null;
        }

        return fields[pos];
    }

    /**
     * @param line
     * @return
     */
    private String[] getFields(final String line) {
        final StringTokenizer izer = new StringTokenizer(line, fieldDelimiters);
        final int size = izer.countTokens();
        final String[] fields = new String[size];

        for (int idx = 0; idx < size; idx++) {
            fields[idx] = izer.nextToken();
        }
        return fields;
    }

    private long getTimeStamp(final String[] fields) {
        if (timeStampPos < 0 || timeStampFormat == null) {
            return -1;
        }

        final DateFormat format = new SimpleDateFormat(timeStampFormat);
        final String timeStamp = fields[timeStampPos];
        Date parsed;

        try {
            parsed = format.parse(timeStamp);
        } catch (final java.text.ParseException e) {
            throw new ParseException(e);
        }

        return parsed.getTime();
    }

    @Override
    public void parse(final InputStream in, final LogListener listener) throws IOException, ParseException {
        final BufferedReader reader = new BufferedReader(new InputStreamReader(in));

        for (;;) {
            final String line = reader.readLine();

            if (line == null) {
                break;
            }

            parseLine(line, listener);
        }
    }

    private void parseLine(final String line, final LogListener listener) {
        final String[] fields = getFields(line);

        if (maxPos >= fields.length) {
            listener.junk(line);
            return;
        }

        final long timestamp = getTimeStamp(fields);
        final String clazz = getField(fields, classPos);
        final String method = getField(fields, methodPos);
        final String level = getField(fields, levelPos);

        listener.recognized(timestamp, clazz, method, level, clazz + '#' + method, fields);
    }
}
