/**
 * Copyright (c) 2010 jolira. All rights reserved. This program and the accompanying materials are made available under
 * the terms of the GNU Public License 2.0 which is available at http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 */

package com.google.code.joliratools.logdb.parser;

import java.io.IOException;
import java.io.InputStream;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This is an immutable class for efficiently parsing log files given a set of regular expressions. This class is
 * thread-safe.
 * 
 * @author ikrimer
 * @since 1.0
 */
public class LogParser implements Parser {
    // if different instances of this class can have different patterns and formats remove static
    private static Pattern headerPattern;
    private static Collection<Pattern> patternsWithFlags = new LinkedHashSet<Pattern>();

    // TODO move following into a config file
    private static final int capacity = 1024 << 2; // should be multiple of the page size
    private static final int bufferSize = 1024 << 2; // should be multiple of the page size
    private static final SimpleDateFormat formatter = new SimpleDateFormat("MMM dd, yyyy hh:mm:ss aa");
    private static final String[] empty = {};

    /**
     * Create a new parser from a set of regular expression describing the patterns to be evaluated in a log file.
     * 
     * @param regex
     *            the map of regular expression to be used for parsing to regex flags CASE_INSENSITIVE, MULTILINE,
     *            DOTALL, UNICODE_CASE, and CANON_EQ The map is linked hash map to emphasize relevance of order and
     *            uniqueness
     */
    public LogParser(final LinkedHashMap<String, Integer> regex) {
        final Iterator<String> patternIterator = regex.keySet().iterator();
        String headerPatternString = patternIterator.next();
        final Integer flags = regex.get(headerPatternString);
        final int headerPatternFlags = flags.intValue();
        // lazy initialize patterns
        if (null == headerPattern || null == patternsWithFlags) {
            headerPattern = Pattern.compile(headerPatternString, headerPatternFlags);
            while (patternIterator.hasNext()) {
                headerPatternString = patternIterator.next();
                patternsWithFlags.add(Pattern.compile(headerPatternString, headerPatternFlags));
            }
        }
    }

    private void doMatching(final CharSequence lines, final List<String> headerFields, final LogListener listener)
            throws java.text.ParseException {
        for (final Pattern p : patternsWithFlags) {
            final Matcher matcher = p.matcher(lines);
            if (matcher.find()) {
                final int numberOfGroups = matcher.groupCount();
                String[] vars;
                vars = numberOfGroups > 0 ? new String[numberOfGroups + 1] : empty;
                for (int i = 0; i < numberOfGroups; i++) {
                    vars[i] = matcher.group(i);
                }
                listener.recognized(formatter.parse(headerFields.get(0)).getTime(), headerFields.get(1),
                        headerFields.get(2), headerFields.get(3), p.pattern(), vars);
            } else {
                listener.unrecognized(formatter.parse(headerFields.get(0)).getTime(), headerFields.get(1),
                        headerFields.get(2), headerFields.get(3), lines.toString());
            }
        }
    }

    private void match(final CharSequence lines, final List<String> headerGroups, final LogListener listener) {
        try {
            doMatching(lines, headerGroups, listener);
        } catch (final java.text.ParseException e) {
            throw new ParseException(e);
        }
    }

    /**
     * Called to parse a log file.
     * 
     * @param in
     *            the input stream the log file should be read from
     * @param listener
     *            the listener to call when records where found
     * @throws IOException
     *             io exception
     * @throws ParseException
     *             any parsing exception
     */
    @Override
    public void parse(final InputStream in, final LogListener listener) throws IOException {
        // TODO if header is fixed replace regex with jflex
        final StringBuilder lines = new StringBuilder(capacity);
        Matcher matcher;
        final List<String> headerGroups = new LinkedList<String>();
        final byte[] buffer = new byte[bufferSize];
        int byteCount;

        int start, end;

        while (0 < (byteCount = in.read(buffer))) {
            lines.append(new String(buffer, 0, byteCount - 1));
            matcher = headerPattern.matcher(lines.substring(0));
            while (0 < lines.length() && matcher.find()) {
                start = matcher.start();
                end = matcher.end();

                final String _lines = lines.substring(0, start);

                if (0 < headerGroups.size()) {
                    match(_lines, headerGroups, listener);
                    headerGroups.clear();
                } else if (0 < start) {// there is no previous header i.e. the line is junk
                    listener.junk(_lines);
                }

                // TODO if header is configurable number of groups should be configurable as well
                for (int i = 1; i <= 4; i++) {
                    headerGroups.add(matcher.group(i));
                }

                // cleanup
                lines.delete(0, end - 1);
                lines.trimToSize();
                matcher = headerPattern.matcher(lines.substring(0));
            }
        }

        // record last line
        if (0 < lines.length()) {
            match(lines, headerGroups, listener);
        }
    }
}
